/*
 * U-Boot and jz4740 LCD controller definitions.
 *
 * Copyright (C) 2001 Wolfgang Denk, DENX Software Engineering, wd@denx.de.
 * Copyright (C) 2005-2007, Ingenic Semiconductor Inc.
 * Copyright (C) 2009 Qi Hardware Inc.
 * Author: Xiangfu Liu <xiangfu@sharism.cc>
 * Copyright (C) Xiangfu Liu <xiangfu@sharism.cc>
 * Copyright (C) 2015, 2016, 2017, 2018 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#ifndef __JZLCD_H__
#define __JZLCD_H__

#include <stdint.h>

/* Framebuffer characteristics. */

struct jzfb_info {
	uint32_t cfg;	/* panel mode and pin usage etc. */
	uint32_t w;	/* display width in pixels */
	uint32_t h;	/* display height in pixels */
	uint32_t bpp;	/* bits per pixel */
	uint32_t fclk;	/* frame clock */
	uint32_t hsw;	/* hsync width, in pixel clock */
	uint32_t vsw;	/* vsync width, in line count */
	uint32_t elw;	/* end of line, in pixel clock */
	uint32_t blw;	/* begin of line, in pixel clock */
	uint32_t efw;	/* end of frame, in line count */
	uint32_t bfw;	/* begin of frame, in line count */
};

/* LCD controller stucture for jz4740. */

struct jz_fb_dma_descriptor {
	struct jz_fb_dma_descriptor *fdadr;	/* frame descriptor address register */
	uint32_t fsadr;				/* frame source address register */
	uint32_t fidr;				/* frame identifier register */
	uint32_t ldcmd;				/* command register */
};

/* Display characteristics and memory resources. */

typedef struct vidinfo {
	struct jzfb_info *jz_fb;		/* framebuffer and panel properties */
	void *lcd;				/* address of LCD controller registers */
} vidinfo_t;



/* Public functions. */

uint32_t jz4740_lcd_get_aligned_size(vidinfo_t *vid);
uint32_t jz4740_lcd_get_descriptors_size(vidinfo_t *vid);
uint32_t jz4740_lcd_get_line_size(vidinfo_t *vid);
uint32_t jz4740_lcd_get_screen_size(vidinfo_t *vid);
uint32_t jz4740_lcd_get_pixel_clock(vidinfo_t *vid);

void jz4740_lcd_set_base(vidinfo_t *vid, void *lcd_base);

void jz4740_lcd_ctrl_init(
	struct jz_fb_dma_descriptor *desc_vaddr,
	struct jz_fb_dma_descriptor *desc_paddr,
	void *fb_vaddr, void *fb_paddr,
	vidinfo_t *vid);

void jz4740_lcd_hw_init(vidinfo_t *vid);
void jz4740_lcd_set_bpp(uint8_t bpp, vidinfo_t *vid);
void jz4740_lcd_enable(vidinfo_t *vid);
void jz4740_lcd_disable(vidinfo_t *vid);
void jz4740_lcd_quick_disable(vidinfo_t *vid);



/* Alignment/rounding macros. */

#define ALIGN(x,a)              __ALIGN_MASK((x),(typeof(x))(a)-1)
#define __ALIGN_MASK(x,mask)    (((x)+(mask))&~(mask))

/* Display device mode select (LCD_CFG.MODE). */

#define MODE_MASK		0x0f
#define MODE_TFT_GEN		0x00
#define MODE_TFT_SHARP		0x01
#define MODE_TFT_CASIO		0x02
#define MODE_TFT_SAMSUNG	0x03
#define MODE_CCIR656_NONINT	0x04
#define MODE_CCIR656_INT	0x05
#define MODE_STN_COLOR_SINGLE	0x08
#define MODE_STN_MONO_SINGLE	0x09
#define MODE_STN_COLOR_DUAL	0x0a
#define MODE_STN_MONO_DUAL	0x0b
#define MODE_8BIT_SERIAL_TFT    0x0c

/* 16-bit or 18-bit TFT panel selection (LCD_CFG.18/16). */

#define MODE_TFT_18BIT          (1<<7)

/* STN pin utilisation (LCD_CFG.PDW). */

#define STN_DAT_PIN1		(0x00 << 4)
#define STN_DAT_PIN2		(0x01 << 4)
#define STN_DAT_PIN4		(0x02 << 4)
#define STN_DAT_PIN8		(0x03 << 4)
#define STN_DAT_PINMASK		STN_DAT_PIN8

/* Pin reset states (LCD_CFG). */

#define STFT_PSHI		(1 << 15)
#define STFT_CLSHI		(1 << 14)
#define STFT_SPLHI		(1 << 13)
#define STFT_REVHI		(1 << 12)

/* Sync direction (LCD_CFG.SYNDIR). */

#define SYNC_MASTER		(0 << 16)
#define SYNC_SLAVE		(1 << 16)

/* Data enable polarity (LCD_CFG.DEP). */

#define DE_P			(0 << 9)
#define DE_N			(1 << 9)

/* Pixel clock polarity (LCD_CFG.PCP). */

#define PCLK_P			(0 << 10)
#define PCLK_N			(1 << 10)

/* Horizontal sync polarity (LCD_CFG.HSP). */

#define HSYNC_P			(0 << 11)
#define HSYNC_N			(1 << 11)

/* Vertical sync polarity (LCD_CFG.VSP). */

#define VSYNC_P			(0 << 8)
#define VSYNC_N			(1 << 8)

/* Inverse output data (LCD_CFG.INVDAT). */

#define DATA_NORMAL		(0 << 17)
#define DATA_INVERSE		(1 << 17)

/* Register offsets. */

#define LCD_CFG			0x00 /* LCD Configure Register */
#define LCD_VSYNC		0x04 /* Vertical Synchronize Register */
#define LCD_HSYNC		0x08 /* Horizontal Synchronize Register */
#define LCD_VAT			0x0c /* Virtual Area Setting Register */
#define LCD_DAH			0x10 /* Display Area Horizontal Start/End Point */
#define LCD_DAV			0x14 /* Display Area Vertical Start/End Point */
#define LCD_PS			0x18 /* PS Signal Setting */
#define LCD_CLS			0x1c /* CLS Signal Setting */
#define LCD_SPL			0x20 /* SPL Signal Setting */
#define LCD_REV			0x24 /* REV Signal Setting */
#define LCD_CTRL		0x30 /* LCD Control Register */
#define LCD_STATE		0x34 /* LCD Status Register */
#define LCD_IID			0x38 /* Interrupt ID Register */
#define LCD_DA0			0x40 /* Descriptor Address Register 0 */
#define LCD_SA0			0x44 /* Source Address Register 0 */
#define LCD_FID0		0x48 /* Frame ID Register 0 */
#define LCD_CMD0		0x4c /* DMA Command Register 0 */
#define LCD_DA1			0x50 /* Descriptor Address Register 1 */
#define LCD_SA1			0x54 /* Source Address Register 1 */
#define LCD_FID1		0x58 /* Frame ID Register 1 */
#define LCD_CMD1		0x5c /* DMA Command Register 1 */

/* Burst length selection (LCD_CTRL.BST). */

#define LCD_CTRL_BST_MASK	(0x03 << 28)
#define LCD_CTRL_BST_4		(0 << 28) /* 4-word */
#define LCD_CTRL_BST_8		(1 << 28) /* 8-word */
#define LCD_CTRL_BST_16		(2 << 28) /* 16-word */

/* RGB mode (LCD_CTRL.RGB). */

#define LCD_CTRL_RGB565		(0 << 27) /* RGB565 mode */
#define LCD_CTRL_RGB555		(1 << 27) /* RGB555 mode */

/* Output FIFO underrun protection (LCD_CTRL.OFUP). */

#define LCD_CTRL_OFUP		(1 << 26) /* Output FIFO underrun protection enable */

/* STN FRC algorithm selection (LCD_CTRL.FRC). */

#define LCD_CTRL_FRC_16		(0 << 24) /* 16 grayscale */
#define LCD_CTRL_FRC_4		(1 << 24) /* 4 grayscale */
#define LCD_CTRL_FRC_2		(2 << 24) /* 2 grayscale */
#define LCD_CTRL_FRC_MASK	(0x03 << 24)

/* Load palette delay counter (LCD_CTRL.PDD) */

#define LCD_CTRL_PDD_MASK	(0xff << 16)

#define LCD_CTRL_EOFM		(1 << 13) /* EOF interrupt mask */
#define LCD_CTRL_SOFM		(1 << 12) /* SOF interrupt mask */
#define LCD_CTRL_OFUM		(1 << 11) /* Output FIFO underrun interrupt mask */
#define LCD_CTRL_IFUM0		(1 << 10) /* Input FIFO 0 underrun interrupt mask */
#define LCD_CTRL_IFUM1		(1 << 9)  /* Input FIFO 1 underrun interrupt mask */
#define LCD_CTRL_LDDM		(1 << 8)  /* LCD disable done interrupt mask */
#define LCD_CTRL_QDM		(1 << 7)  /* LCD quick disable done interrupt mask */
#define LCD_CTRL_BEDN		(1 << 6)  /* Endian selection */
#define LCD_CTRL_PEDN		(1 << 5)  /* Endian in byte:0-msb first, 1-lsb first */
#define LCD_CTRL_DIS		(1 << 4)  /* Disable indicate bit */
#define LCD_CTRL_ENA		(1 << 3)  /* LCD enable bit */

/* Bits per pixel (LCD_CTRL.BPP). */

#define LCD_CTRL_BPP_1		0 /* 1 bpp */
#define LCD_CTRL_BPP_2		1 /* 2 bpp */
#define LCD_CTRL_BPP_4		2 /* 4 bpp */
#define LCD_CTRL_BPP_8		3 /* 8 bpp */
#define LCD_CTRL_BPP_16		4 /* 15/16 bpp */
#define LCD_CTRL_BPP_18_24	5 /* 18/24/32 bpp */
#define LCD_CTRL_BPP_MASK	0x07

/* Palette buffer (LCD_CMDx.PAL). */

#define LCD_CMD_PAL		(1 << 28)

#endif /* __JZLCD_H__ */
